from typing import Coroutine, Optional
from ..world import AutomationState
from .. import event
from . import Actor


class Automator(Actor):
    interval: float = 0.1

    def __init__(self, default_automation: Coroutine = None) -> None:
        super().__init__()
        self.routines: list[Coroutine] = []
        self.default_automation = default_automation
        event.register(event.Id.PAUSE_AUTOMATIONS, self._pause)

    def add(self, coro: Coroutine):
        self.routines.append(coro)

    def replace(self, coro: Coroutine):
        self.clear()
        self.add(coro)

    def clear(self):
        '''clears all automations'''
        [r.close() for r in self.routines]  # NOTE: this ensures we do not get warnings about missing await for our routines
        self.routines.clear()

    async def step(self):
        if not self.routines or self.world.automation_state == AutomationState.DISABLED:
            if not self.routines:
                if self.default_automation:
                    self.log.info('automations were disabled, now using default automation')
                    self.add(self.default_automation())
                else:
                    self.world.automation_state = AutomationState.DISABLED
            if self.routines:
                self.world.automation_state = AutomationState.STOPPED

        if self.world.automation_state != AutomationState.RUNNING:
            return

        for coro in self.routines:
            try:
                coro.send(None)
            except StopIteration:
                self.routines.remove(coro)
                if not self.routines:
                    await self.pause_automations(because='the last one has completed')
                    self.world.automation_state = AutomationState.DISABLED
            except:
                await self.pause_automations(because='an exception occurred in an automation')
                self.routines.clear()
                self.log.exception(f'paused and cleared automations due to exception in {coro}')
                self.world.automation_state = AutomationState.DISABLED

    async def _pause(self, because: Optional[str] = None):
        '''Pauses the automation.

        Only to be used internally.
        The proper way is to use runtime.pause(...) or fire event.Id.PAUSE_AUTOMATION.
        See rosys.io/automations.
        '''
        if self.world.automation_state == AutomationState.PAUSED:
            return
        self.world.automation_state = AutomationState.PAUSED
        if because:
            await event.call(event.Id.NEW_NOTIFICATION, f'pausing automations because {because}')

    async def tear_down(self):
        await super().tear_down()
        self.clear()



from .. import event
from ..hardware import Hardware
from . import Actor


class Lizard(Actor):
    interval: float = 0.01

    def __init__(self, hardware: Hardware):
        super().__init__()
        self.hardware = hardware
        event.register(event.Id.PAUSE_AUTOMATIONS, self._handle_pause)

    async def step(self):
        await self.hardware.update()
        await event.call(event.Id.NEW_MACHINE_DATA)

    async def _handle_pause(self, reason: str):
        await self.hardware.stop()



from typing import Optional
import abc


class Communication(abc.ABC):

    @classmethod
    @abc.abstractmethod
    def is_possible(cls) -> bool:
        return False

    @abc.abstractmethod
    async def send_async(self, msg: str):
        pass

    @abc.abstractmethod
    async def read(self) -> Optional[str]:
        pass



from typing import Optional
import asyncio
import os
import serial
from .checksum import augment, check
from .communication import Communication


class SerialCommunication(Communication):
    device_path: str = '/dev/esp'
    baudrate: int = 115200

    def __init__(self):
        self.serial = serial.Serial(self.device_path, self.baudrate)
        self.buffer = ''

    @classmethod
    def is_possible(cls) -> bool:
        return os.path.exists(cls.device_path) and os.stat(cls.device_path).st_gid > 0

    def connect(self):
        if not self.serial.isOpen():
            self.serial.open()

    def disconnect(self):
        if self.serial.isOpen():
            self.serial.close()

    async def read(self) -> Optional[str]:
        if not self.serial.isOpen():
            return
        self.buffer += self.serial.read_all().decode()
        if '\n' in self.buffer:
            line, self.buffer = self.buffer.split('\r\n', 1)
            return check(line)

    async def send_async(self, line: str):
        self.serial.write(f'{augment(line)}\n'.encode())
        await asyncio.sleep(0)



from pydantic import BaseModel
from typing import Optional
import asyncio
import numpy as np
from .communication import Communication, CommunicationFactory
from .world import Mode, Velocity, World


class Simulation(BaseModel):
    linear_velocity: float = 0
    angular_velocity: float = 0


class Hardware:

    def __init__(self, world: World, communication: Optional[Communication] = ...):
        self.world = world
        self.simulation = Simulation()
        if communication is ...:
            self.communication = CommunicationFactory.create() if world.mode == Mode.REAL else None
        else:
            self.communication = communication

    async def configure(self, filepath: str):
        if self.communication:
            with open(filepath) as f:
                expander = False
                await self.communication.send_async(f'!-')
                for line in f.read().splitlines():
                    if line == '---':
                        expander = True
                        await self.communication.send_async('!>!-')
                    else:
                        if expander:
                            await self.communication.send_async(f'!>!+{line}')
                        else:
                            await self.communication.send_async(f'!+{line}')
                if expander:
                    await self.communication.send_async(f'!>!.')
                    await self.communication.send_async(f'!>core.restart()')
                await self.communication.send_async(f'!.')
                await self.restart()
        else:
            raise Exception('there is no communication for configuring hardware')

    async def restart(self):
        if self.communication:
            await self.communication.send_async(f'core.restart()')
        else:
            raise Exception('there is no communication for restarting hardware')

    async def drive(self, linear: float, angular: float):
        if self.communication:
            await self.communication.send_async(f'wheels.speed({linear}, {angular})')
        else:
            self.simulation.linear_velocity = linear
            self.simulation.angular_velocity = angular
            await asyncio.sleep(0)

    async def stop(self):
        if self.communication:
            await self.communication.send_async('wheels.off()')
        else:
            self.simulation.linear_velocity = 0
            self.simulation.angular_velocity = 0
            await asyncio.sleep(0)

    async def update(self):
        if self.communication:
            millis = None
            while True:
                line = await self.communication.read()
                if line is None:
                    break
                words = line.split()
                if not words:
                    continue
                first = words.pop(0)
                if first not in ['core', '!"core']:
                    continue
                millis = float(words.pop(0))
                if self.world.robot.clock_offset is None:
                    continue
                self.world.robot.hardware_time = millis / 1000 + self.world.robot.clock_offset
                self.parse(words)
            if millis is not None:
                self.world.robot.clock_offset = self.world.time - millis / 1000
        else:
            self.simulate()

    def parse(self, words: list[str]):
        self.world.robot.odometry.append(Velocity(
            linear=float(words.pop(0)),
            angular=float(words.pop(0)),
            time=self.world.robot.hardware_time,
        ))

    def simulate(self):
        self.world.robot.odometry.append(Velocity(
            linear=self.simulation.linear_velocity,
            angular=self.simulation.angular_velocity,
            time=self.world.time,
        ))
        self.world.robot.battery = 25.0 + np.sin(0.1 * self.world.time) + 0.02 * np.random.randn()
        self.world.robot.temperature = np.random.uniform(34, 35)



def augment(line: str) -> str:
    checksum = 0
    for c in line:
        checksum ^= ord(c)
    return f'{line}@{checksum:02x}'


def check(line: str) -> str:
    if line[-3:-2] == '@':
        check = int(line[-2:], 16)
        line = line[:-3]
        checksum = 0
        for c in line:
            checksum ^= ord(c)
        if checksum != check:
            return None
    return line



from nicegui.ui import Ui
from .. import Runtime
from ..world.world import AutomationState


class AutomationControls:
    ui: Ui
    runtime: Runtime

    def __init__(self) -> None:

        async def toggle_automation():
            if self.runtime.world.automation_state != AutomationState.RUNNING:
                await self.runtime.resume()
            else:
                await self.runtime.pause(because='pause button was pressed')

        async def stop():
            await self.runtime.stop(because='stop button was pressed')

        automation_button = self.ui.button(on_click=toggle_automation).props('icon=play_arrow outline')
        stop_button = self.ui.button(on_click=stop).props('icon=stop outline')

        def refresh():
            if self.runtime.world.automation_state == AutomationState.DISABLED:
                automation_button.props(replace='icon=play_arrow')
                automation_button.view.disable = stop_button.view.disable = True
            elif self.runtime.world.automation_state == AutomationState.STOPPED:
                automation_button.props(replace='icon=play_arrow')
                automation_button.view.disable = False
                stop_button.view.disable = True
            elif self.runtime.world.automation_state == AutomationState.RUNNING:
                automation_button.props(replace='icon=pause')
                automation_button.view.disable = stop_button.view.disable = False
            elif self.runtime.world.automation_state == AutomationState.PAUSED:
                automation_button.props(replace='icon=play_arrow')
                automation_button.view.disable = stop_button.view.disable = False
        self.ui.timer(0.1, refresh)



from pydantic import BaseModel, PrivateAttr
from aenum import Enum, auto
import time
from .mode import Mode
from .obstacle import Obstacle
from .robot import Robot


class AutomationState(str, Enum, init='value __doc__'):

    def _generate_next_value_(name, start, count, last_values):
        '''uses enum name as value when calling auto()'''
        return name

    DISABLED = auto(), 'no automations available or execution not allowed'
    STOPPED = auto(), 'there is an automation which could be started'
    RUNNING = auto(), 'automations are beeing processed'
    PAUSED = auto(), 'an ongoing automation can be resumed'


class World(BaseModel):

    robot: Robot = Robot()
    mode: Mode = Mode.REAL
    automation_state: AutomationState = AutomationState.DISABLED
    _time: float = PrivateAttr(default_factory=time.time)
    tracking: bool = False
    robot_locator_cam: str = None
    download_queue: list[str] = []
    image_data: dict[str, bytes] = {}
    link_queue: list[list[str]] = []
    obstacles: dict[str, Obstacle] = {}
    notifications: list[tuple[float, str]] = []

    @property
    def time(self):
        return self._time if self.mode == Mode.TEST else time.time()

    def set_time(self, value):
        assert self.mode == Mode.TEST
        self._time = value



from nicegui.ui import Ui
import starlette
from rosys import Runtime


def setup(ui: Ui, runtime: Runtime):
    ui.add_route(starlette.routing.Route(
        '/world', lambda request, **_:
        starlette.responses.Response(content=runtime.world.json(exclude={'image_data'}), media_type='text/json')))
    ui.add_route(starlette.routing.Route(
        '/export', lambda request, **_:
        starlette.responses.JSONResponse(content=runtime.persistence.dump())))



from __future__ import annotations
from asyncio.exceptions import CancelledError
import asyncio
import logging
from typing import Optional, Type
from . import event, task_logger
from .actors import Actor, Automator, Lizard, Odometer, Steerer
from .hardware import Hardware
from .persistence import Persistence
from .world import AutomationState, Mode, World


class Runtime:

    def __init__(self,
                 world: Optional[World] = None,
                 persistence: Optional[Persistence] = None,
                 hardware: Optional[Hardware] = None):
        self.world = world or World()
        Actor.world = self.world
        self.tasks = []
        self.log = logging.getLogger(__name__)

        if self.world.mode != Mode.TEST:
            self.persistence = persistence or Persistence(self.world)
            self.persistence.restore()

        self.hardware = hardware or Hardware(self.world)
        self.lizard = Lizard(self.hardware)
        self.odometer = Odometer()
        self.steerer = Steerer(self.hardware)
        self.automator = Automator()

        self.actors = [
            self.lizard,
            self.odometer,
            self.steerer,
            self.automator,
        ]

    def with_actors(self, *actors: list[Actor]):
        self.actors += actors
        return self

    async def pause(self, because: Optional[str] = None):
        await event.call(event.Id.PAUSE_AUTOMATIONS, because)

    async def stop(self, because: Optional[str] = None):
        await event.call(event.Id.PAUSE_AUTOMATIONS, because)
        self.automator.routines.clear()
        self.world.automation_state = AutomationState.STOPPED

    async def resume(self):
        previous = self.world.automation_state
        if previous != AutomationState.DISABLED:
            self.world.automation_state = AutomationState.RUNNING
            if previous == AutomationState.STOPPED:
                event.emit(event.Id.AUTOMATIONS_STARTED)

    async def startup(self):
        if self.tasks:
            raise Exception('should be only executed once')

        event.register(event.Id.NEW_NOTIFICATION, self.store_notification)
        for actor in self.actors:
            if actor.interval is not None:
                self.tasks.append(task_logger.create_task(self.repeat(actor), name=actor.name))
        self.tasks.append(asyncio.create_task(self.watch_emitted_events()))

        await asyncio.sleep(1)  # NOTE we wait for RoSys to start up before analyzing async debugging
        self.activate_async_debugging()

    async def shutdown(self):
        await self.hardware.drive(0, 0)
        if self.world.mode != Mode.TEST:
            self.persistence.backup()
        [t.cancel() for t in self.tasks]
        Actor.process_pool.shutdown()
        await asyncio.gather(*[task_logger.create_task(a.tear_down(), name=f'{a.name}.tear_down()') for a in self.actors])

    async def repeat(self, actor: Actor):
        while True:
            start = self.world.time
            try:
                await actor.step()
                dt = self.world.time - start
            except (CancelledError, GeneratorExit):
                return
            except:
                dt = self.world.time - start
                self.log.exception(f'error in {actor}')
                if actor.interval == 0 and dt < 0.1:
                    delay = 0.1 - dt
                    self.log.warning(
                        f'{actor} would be called to frequently ' +
                        f'because it only took {dt*1000:.0f} ms; ' +
                        f'delaying this step for {delay*1000:.0f} ms')
                    await actor.sleep(delay)
            try:
                await actor.sleep(actor.interval - dt)
            except (CancelledError, GeneratorExit):
                return

    def get_actor(self, _type: Type):
        return next((a for a in self.actors if type(a) is _type), None)

    def store_notification(self, message: str):
        self.log.info(message)
        self.world.notifications.append((self.world.time, message))

    def activate_async_debugging(self):
        '''Produce warnings for coroutines which take too long on the main loop and hence clog the event loop'''
        try:
            loop = asyncio.get_running_loop()
            loop.set_debug(True)
            loop.slow_callback_duration = 0.05
        except:
            self.log.exception('could not activate async debugging')

    async def watch_emitted_events(self):
        while True:
            try:
                for task in event.tasks:
                    if task.done() and task.exception():
                        self.handle_exception(task.exception())
                # cleanup finished tasks
                event.tasks = [t for t in event.tasks if not t.done()]
            except:
                self.log.exception('failing to watch emitted events')
            await asyncio.sleep(0 if self.world.mode == Mode.TEST else 0.1)

    def handle_exception(self, ex: Exception):
        self.log.exception('task failed to execute', exc_info=ex)



from .mode import Mode
from .obstacle import Obstacle
from .path_segment import PathSegment
from .point import Point
from .point3d import Point3d
from .pose import Pose, PoseStep
from .robot import Robot, RobotParameters, RobotShape
from .rotation import Rotation
from .spline import Spline
from .velocity import Velocity
from .world import World, AutomationState
